-- opb_la module
-- internal logic state analyser
-- this device uses only 8-bit OPB bus - c_opb_dwidth should be only 8 - the rest LSBs are ignored
-- use opb2opb_dwidth component to convert data width when other data width is used

-- by Erenst Jamro 
--//////////////////////////////////////////////////////////////////////
--//// Copyright (C) 2003 Authors                                   //// 
--////                                                              ////
--//// This source file may be used and distributed without         ////
--/// restriction provided that this copyright statement is not     ////
--//// removed from the file and that any derivative work contains  ////
--//// the original copyright notice and the associated disclaimer. ////
--////                                                              ////
--//// This source file is free software; you can redistribute it   ////
--//// and/or modify it under the terms of the GNU Lesser General   ////
--//// Public License as published by the Free Software Foundation; ////
--//// either version 2.1 of the License, or (at your option) any   ////
--//// later version.                                               ////
--////                                                              ////
--//// This source is distributed in the hope that it will be       ////
--//// useful, but WITHOUT ANY WARRANTY; without even the implied   ////
--//// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      ////
--//// PURPOSE. See the GNU Lesser General Public License for more  ////
--//// details.                                                     ////

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

library UNISIM;
use unisim.all; -- for global set reset signal

entity opb_la is
  generic (c_opb_dwidth: integer:= 32; -- OPB data bus width
           c_opb_awidth: integer:= 32; -- OPB address bus width
  		   c_baseaddr: std_logic_vector:= X"00000000"; -- la control based address
		   c_highaddr: std_logic_vector:= X"000003FF";
		   c_la_run_length_coding: integer:= 0; -- include run length codding logic
		   c_la_dwidth: integer:= 32; -- internal logic analyser data width= 8, 16, 32 or 64
		   c_la_dtype: integer:= 0; -- 0- use internal data, 1- use external data, 2- OPB_ABus, 3- OPB_DBus
           c_la_twidth: integer:= 8; -- trigger width
		   c_la_ttype: integer:= 0; -- trigger data type= 0- internal, 1- external
		   c_la_ce_dwidth: integer:= 8; -- width of the data clock enable
		   c_la_ce_dtype: integer:= 3; -- data clock enable= 0- standard connecttion, 1 - external, 2- same as data, 3- same as trigger
		   c_la_ce_ttype: integer:= 0; -- type of the clock enable for trigger
		   c_la_two_clk: integer:= 0); -- use two different clock for control and capture interface

  port (OPB_clk: in STD_LOGIC; 
		-- OPB Slave interface - Control interface
       SOPB_ABus    : in std_logic_vector(0 to C_OPB_AWIDTH-1);
       SOPB_BE      : in std_logic_vector(0 to C_OPB_DWIDTH/8-1);
       SOPB_RNW     : in std_logic;
       SOPB_select  : in std_logic;
       SOPB_seqAddr : in std_logic;
       SOPB_DBus    : in std_logic_vector(0 to C_OPB_DWIDTH-1);
       Sl_DBus    : out std_logic_vector(0 to C_OPB_DWIDTH-1);
       Sl_errAck  : out std_logic;
       Sl_retry   : out std_logic;
       Sl_toutSup : out std_logic;
       Sl_xferAck : out std_logic;

       -- OPB Master interface - data capture interface
       M_ABus : out STD_LOGIC_VECTOR(0 TO c_opb_awidth-1); -- always zero
	   M_DBus : out STD_LOGIC_VECTOR(0 TO c_opb_dwidth-1); -- always zero
       M_BE : out STD_LOGIC_VECTOR(0 TO c_opb_dwidth/8 -1); -- always zero
	   M_RNW : out STD_LOGIC; -- always zero
	   M_select : out STD_LOGIC; -- always zero
	   M_seqAddr : out STD_LOGIC; -- always zero
	   M_busLock : out STD_LOGIC; -- always zero
	   M_request : out STD_LOGIC;  -- always zero 

	   MOPB_MGrant: in STD_LOGIC; -- open
	   MOPB_pendReq: in STD_LOGIC;

	   OPB_rst: in STD_LOGIC;
	   MOPB_Select: in STD_LOGIC;
       MOPB_BE      : in std_logic_vector(0 to C_OPB_DWIDTH/8-1);
       MOPB_RNW     : in std_logic;
       MOPB_seqAddr : in std_logic;
       MOPB_busLock : in std_logic;
	   MOPB_ABus : in STD_LOGIC_VECTOR(0 TO c_opb_awidth-1);
	   
	   MOPB_DBus : in STD_LOGIC_VECTOR(0 TO c_opb_dwidth-1);
	   MOPB_errAck : in STD_LOGIC;
	   MOPB_retry : in STD_LOGIC;
	   MOPB_timeout: in STD_LOGIC;
	   MOPB_xferAck : in STD_LOGIC;
	   la_DBus: in std_logic_vector(0 to c_la_dwidth-1);
	   la_TBus: in std_logic_vector(0 to c_la_twidth-1);
	   la_ceDBus: in std_logic_vector(0 to c_la_ce_dwidth-1); -- clock enable for la_DBus	
	   la_ceTBus: in std_logic
	);
end opb_la;

architecture opb_la_arch of opb_la is

component ROC -- reset on Configuration
  port (O : out std_logic);
  end component;

component OPB_slave
  generic (
    C_OPB_AWIDTH : integer                   := 32;
    C_OPB_DWIDTH : integer                   := 32;
    C_BASEADDR   : std_logic_vector          := X"0000_0100";
    C_HIGHADDR   : std_logic_vector          := X"0000_01FF";
	-- user interface data and address width
	DWIDTH : integer:= 8; -- real width of the data - the rest LSBs are always 0s
	AWIDTH : integer:= 8 -- user address width
    );
  port (
    -- OPB signals
    OPB_ABus    : in std_logic_vector(0 to C_OPB_AWIDTH-1);
    OPB_BE      : in std_logic_vector(0 to C_OPB_DWIDTH/8-1);
    OPB_RNW     : in std_logic;
    OPB_select  : in std_logic;
    OPB_seqAddr : in std_logic;
    OPB_DBus    : in std_logic_vector(0 to C_OPB_DWIDTH-1);
    Sl_DBus    : out std_logic_vector(0 to C_OPB_DWIDTH-1);
    Sl_errAck  : out std_logic;
    Sl_retry   : out std_logic;
    Sl_toutSup : out std_logic;
    Sl_xferAck : out std_logic;
	-- user signals
	sel: out std_logic; -- select - active when OPB_select and proper address
	ack: in std_logic; -- acknowlenge - slave is ready for data transfer
	rnw: out std_logic; -- read not write - same assert OPB_read not write
	adr: out std_logic_vector(0 to awidth-1); -- address bus
	drd: in std_logic_vector(0 to dwidth-1); -- data read by opb master (send by the uer device)
	dwr: out std_logic_vector(0 to dwidth-1); -- data written by the master (recieved by the user device)
	be: out std_logic_vector(0 to dwidth/8 -1); -- byte enable 
	seqAddr: out std_logic -- same as OPB_seqAddr
    );
  end component;

component log_anal
  generic(data_width: integer:= 16; -- width of the data that are analysed (must be power of 2)
      mem_adr_width: integer:= 9; -- internal memory address width
	  adr_width: integer:= 11; -- adr_I address width
      trig_width: integer:= 8; -- width of the triger logic
	  ce_dwidth: integer:= 1; -- data clock enable trigger-like logic width
	  ced_type: integer:= 1; -- 0 - data clock enable logic is desabled, 1- enable data clock enable logic
	  two_clocks: integer:= 0; -- two seperate clocks for control interface and logic analyser
      use_run_length_coding: integer:= 0);
  -- the following rules must!!! be satysfied for generic:
  -- data_width= 8, 16, 32, 64
  -- adr_width= mem_adr_width + 1 + log2(data_width/8); 4<=mem_adr_width<=16
  -- mem_adr_width<=16
  -- trig_width<= 64
  -- two_clocks= 0 (only wb_clk_I clock is used) 1- f_clk= k*f_wb_clk_I, 2- two asynchronous clocks
  
  -- mem_adr_width optimal selection for 4kb BRAM (Virtex): 
  --  mem_adr_width<= 9 then number of BRAM block used= data_width/8
    
  port (arst: in std_logic; -- global asynchronous set reset signal (mainly for simulation purposes)
	  -- interface for logic analyser 
	  clk: in std_logic; -- seperate clock for logic analyzer (the same as wb_clk_I when two_clocks= 0)
	  data: in std_logic_vector(data_width-1 downto 0); -- data that are analysied
	  ce_data: in std_logic_vector(ce_dwidth-1 downto 0); -- clock enable -- should be used if data are recorded e.g. every second la_clk
      trig: in std_logic_vector(trig_width-1 downto 0); -- triger bus (can be different that data bus
	  ce_trig: in std_logic; -- clock enable for triger bus
	  -- control WISHBOBE slave interface - interface for setting logic analyser options and transfering analysed data to computer
      wb_clk_I: in STD_LOGIC; -- clock (common for every logic every logic analyser signals) - common for loagic analyser and WISHBONE control interface
	  wb_adr_I: in std_logic_vector(adr_width-1 downto 0); -- address bus (one bit wider than mem_adr_width)
	  wb_dat_I: in std_logic_vector(7 downto 0);
	  wb_dat_O: out std_logic_vector(7 downto 0);
	  wb_stb_I, wb_we_I: in std_logic; 
	  -- wb_cyc_I: in std_logic; -- signal is ignored
	  -- wb_rst_I: in std_logic; -- the WISHBONE interface need not be reseted
	  wb_ack_O: buffer std_logic); 
  end component;
	
  function log2(arg: integer) return integer is
    variable ret: integer:= -1;
  begin
   for i in 0 to 32 loop
	  if 2**i = arg then
		  ret:= i;
		  return i;
	  end if;
	end loop;
	assert false -- always assert when loop does not finish the function
	report "Address space is not a power of 2"
	severity failure;
	return -1;
end log2;
  constant adr_space: integer:= conv_integer(c_highaddr - c_baseaddr + 1);
  constant awidth: integer:= log2(adr_space); -- address width of the LA
  constant mem_awidth: integer:= awidth - 1 - log2(c_la_dwidth/8); -- internal memory address width
  
  signal arst: std_logic; -- asynchronous global reset
  signal adr: std_logic_vector(0 to awidth-1); -- slave address
  signal sel, ack, rnw, we: std_logic; -- logic analyzer module
  signal drd, dwr: std_logic_vector(7 downto 0); -- control interface data for read and write
  
  -- LA signals
  signal la_data: std_logic_vector(c_la_dwidth-1 downto 0);
  signal la_trig: std_logic_vector(c_la_twidth-1 downto 0);
  signal la_ce_trig: std_logic;
  signal la_std_data: std_logic_vector(63 downto 0); -- standard (internal) LA watch signals
  signal la_std_ce: std_logic_vector(0 to 4); -- clock enable standard signals
begin
  -- opb_slave module
  sl: OPB_slave
  generic map (C_OPB_AWIDTH=> c_opb_awidth, C_OPB_DWIDTH=> c_opb_dwidth,
    C_BASEADDR=> c_baseaddr, C_HIGHADDR=> c_highaddr,
	-- user interface data and address width
	DWIDTH=> 8, AWIDTH=> awidth)
  port map(OPB_ABus=> SOPB_ABus, OPB_BE=> SOPB_BE, 
    OPB_RNW=> SOPB_RNW, OPB_select=> SOPB_Select,
    OPB_seqAddr=> SOPB_seqAddr, OPB_DBus=> SOPB_DBus, 
    Sl_DBus=> Sl_DBus, Sl_errAck=> Sl_errAck, 
    Sl_retry=> Sl_retry, Sl_toutSup=> Sl_toutSup,
    Sl_xferAck=> Sl_xferAck,
	-- user signals
	sel=> sel, ack=> ack, rnw=> rnw,
	adr=> adr, drd=> drd, dwr=> dwr, be=> open, seqAddr=> open);	
	
	
-----------------------------------------------------------------
-- Master signals not used

  M_ABus<= (others=> '0');
  M_DBus<= (others=> '0');
  M_BE<= (others=> '0');
  M_RNW<= '0';
  M_select<= '0';
  M_seqAddr<= '0';
  M_busLock<= '0';
  M_request<= '0';

-----------------------------------------------------------------
  la: log_anal
      generic map (data_width=> c_la_dwidth, mem_adr_width=> mem_awidth, 
        adr_width=> awidth, trig_width=> c_la_twidth, two_clocks=> c_la_two_clk, 
	    ce_dwidth=> c_la_ce_dwidth, ced_type=> c_la_ce_dtype,
		use_run_length_coding=> c_la_run_length_coding)
      port map (arst=> arst, clk=> OPB_clk,
        -- interface for logic analyser 
	    data=> la_data, ce_data=> la_ceDBus,
        trig=> la_trig, ce_trig=> la_ce_trig,
	    -- control WISHBOBE slave interface - interface for setting logic analyser options and transfering analysed data to computer
        wb_clk_I=> OPB_clk, 
	    wb_adr_I=> adr, wb_dat_I=> dwr, wb_dat_O=> drd,
	    wb_stb_I=> sel, wb_we_I=> we, wb_ack_O=> ack); 
	we<= not rnw;
    -- standard internal signals to be analysed by the LA
	  la_std_data(7 downto 0)<= MOPB_busLock & MOPB_seqAddr & MOPB_RNW & MOPB_errAck 
	     & MOPB_timeout & OPB_rst & MOPB_xferAck & MOPB_select;
	 g32: if c_la_dwidth<=32 generate
      g15: if c_la_dwidth=16 and c_la_run_length_coding=1 generate
        la_std_data(14 downto 8)<= MOPB_DBus(0 to 6);
      end generate;
      gn15: if c_la_dwidth/=16 or c_la_run_length_coding=0 generate
        la_std_data(15 downto 8)<= MOPB_DBus(0 to 7);
      end generate;
      la_std_data(31 downto 16)<= MOPB_ABus(c_opb_awidth-16 to c_opb_awidth-1);
	end generate;
	g64: if c_la_dwidth=64 generate
      la_std_data(39 downto 8)<= MOPB_DBus(0 to 31);
	  la_std_data(63 downto 40)<= MOPB_ABus(c_opb_awidth-24 to c_opb_awidth-1);
	end generate;
   
	-- c_la_dtype (data type)
    dtype0: if c_la_dtype=0 generate -- internal logic
	    la_data<= la_std_data(c_la_dwidth-1 downto 0);
	end generate;
	dtype1: if c_la_dtype = 1 generate -- external data
	   la_data<= la_DBus;
	end generate;
	dtype2: if c_la_dtype = 2 generate -- LSB OPB_ABus
	   la_data<= MOPB_ABus(c_opb_awidth-c_la_dwidth to c_opb_awidth-1); -- LSBs
	end generate;
	dtype3: if c_la_dtype = 3 generate -- MSB OPB_DBus
	   la_data<= MOPB_DBus(0 to c_la_dwidth-1); -- MSBs
	end generate;
    -- c_la_ttype (triger type) -- trigger data type= 0- internal, 1- external, 2- OPB_ABus, 3- OPB_DBus 4- same as la_dbus
	ttype0: if c_la_ttype = 0 generate -- internal (same as internal data)
      la_trig<= la_std_data(c_la_twidth-1 downto 0);
	end generate;
	ttype1: if c_la_ttype = 1 generate -- external
	  la_trig<= la_TBus;
	end generate;
	ttype2: if c_la_ttype = 2 generate -- LSB OPB_ABus
	   la_trig<= MOPB_ABus(c_opb_awidth-c_la_twidth to c_opb_awidth-1); -- LSBs
	end generate;
	ttype3: if c_la_ttype = 3 generate -- MSB OPB_DBus
	   la_trig<= MOPB_DBus(0 to c_la_twidth-1); -- MSBs
	end generate;
	ttype4: if c_la_ttype = 4 generate -- same as la_DBus
	  la_trig<= la_DBus(0 to c_la_twidth-1);
	end generate;

	-----------------------
	-- c_la_tce -- data clock enable= 0- always '1', 1 - external, 2- '1' when OPB_select='1'; 3 - '1' when OPB_xferAck='1', 4 - '1' when OPB_errAck='1' or OPB_xferAck or OPB_retry='1', 5 - OPB_rst
	la_std_ce(0 to 3)<= '1' & la_ceTBus & MOPB_select & MOPB_xferAck;
	la_std_ce(4)<= MOPB_xferAck or MOPB_errAck or MOPB_retry or MOPB_timeout;
    
    la_ce_trig<= la_std_ce(c_la_ce_ttype);
	  
-- global set/reset logic - this component draws rst during chip configuration only (no external reset is used)
 global_reset: ROC
	port map (O=> arst);
	
end opb_la_arch;