-- Author:          Ernest Jamro
-- plik interface urzadzenia slave - kazde urzadzenie slave powinno korzystac z tego modulu

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

entity OPB_slave is
  generic (
    C_OPB_AWIDTH : integer                   := 32;
    C_OPB_DWIDTH : integer                   := 32;
    C_BASEADDR   : std_logic_vector          := X"0000_0100";
    C_HIGHADDR   : std_logic_vector          := X"0000_01FF";
	-- user interface data and address width
	DWIDTH : integer:= 8; -- real width of the data - the rest LSBs are always 0s
	AWIDTH : integer:= 8 -- user address width
    );
  port (
    -- OPB signals
    OPB_ABus    : in std_logic_vector(0 to C_OPB_AWIDTH-1);
    OPB_BE      : in std_logic_vector(0 to C_OPB_DWIDTH/8-1);
    OPB_RNW     : in std_logic;
    OPB_select  : in std_logic;
    OPB_seqAddr : in std_logic;
    OPB_DBus    : in std_logic_vector(0 to C_OPB_DWIDTH-1);
    Sl_DBus    : out std_logic_vector(0 to C_OPB_DWIDTH-1);
    Sl_errAck  : out std_logic;
    Sl_retry   : out std_logic;
    Sl_toutSup : out std_logic;
    Sl_xferAck : out std_logic;
	-- user signals
	sel: out std_logic; -- select - active when OPB_select and proper address
	ack: in std_logic; -- acknowlenge - slave is ready for data transfer
	rnw: out std_logic; -- read not write - same assert OPB_read not write
	adr: out std_logic_vector(0 to awidth-1); -- address bus
	drd: in std_logic_vector(0 to dwidth-1); -- data read by opb master (send by the uer device)
	dwr: out std_logic_vector(0 to dwidth-1); -- data written by the master (recieved by the user device)
	be: out std_logic_vector(0 to dwidth/8 -1); -- byte enable 
	seqAddr: out std_logic -- same as OPB_seqAddr
    );
end OPB_slave;


architecture opb_slave_arch of OPB_slave is
  function log2(arg: integer) return integer is
    variable ret: integer:= -1;
  begin
   for i in 0 to 32 loop
	  if 2**i = arg then
		  ret:= i;
		  return i;
	  end if;
	end loop;
	assert false -- always assert when loop does not finish the function
	report "Address space is not a power of 2"
	severity failure;
	return -1;
end log2;
  constant adr_space: integer:= conv_integer(c_highaddr - c_baseaddr + 1);
  constant this_awidth: integer:= log2(adr_space); -- memory address width

  signal dev_select: std_logic; -- active when OPB_select and device address

begin  
  dev_select<= '1' when OPB_select='1' and
  OPB_ABus(0 to C_OPB_AWIDTH - awidth - 1) 
    = C_BASEADDR(0 to C_OPB_AWIDTH - awidth - 1)
	else '0';
   
  sel<= dev_select;
  
  Sl_xferAck<= ack;
  rnw<= OPB_rnw;
  be<= OPB_BE(0 to (dwidth-1)/8);
  Sl_errAck<= '0';
  Sl_retry<= '0';
  Sl_toutSup<= '0';
  gd: if c_opb_dwidth>dwidth generate -- different data width for OPB_bus and user bus
	 Sl_DBus(dwidth to c_opb_dwidth-1)<= (others=>'0');
  end generate;
  Sl_DBus(0 to dwidth-1)<= drd when dev_select='1' and OPB_rnw='1' else (others=>'0');
  dwr<= OPB_DBus(0 to dwidth-1);
  seqAddr<= OPB_seqAddr;
  adr<= OPB_ABus(c_opb_awidth-awidth to c_opb_awidth-1);

  
  assert this_awidth = awidth
   report "opb_slave module error. Different address width for awidth and address space"
   severity error;

 end opb_slave_arch;